// FLASHUP - EPC System Update Utility (c) 2000 Technologic Systems
//
// Description
// --------------------------
// This program is used to load the onboard Flash, CMOS memory, and RTC 
// values for Technologic Systems EPCs. If the EPC is virgin (never loaded)
// or dead (the BIOS has been trashed), a TS-9100 or TS-9200 board must be 
// used to bootstrap the system. If the EPC is currently functional and
// you are simply updating some or all of the system, then this utility and
// the update image files can be downloaded to the EPC and run directly 
// from there.
//
// Flashup is also used to create the image files from a live EPC, enabling
// you to clone EPC installations.
//
// Flashup can currently handle the following sections:
//      BIOS	- create / restore the BIOS
//	DOS	- create / restore the DOS 
//	DISK	- create / restore the SSD (solid state disk)
//	CMOS	- create / restore the CMOS settings
//	RTC	- copy the time and date from a TS-9200 boot board
//
// Creating Update Images
// --------------------------
// To create the update images, the following syntax is used:
//
// FLASHUP /CREATE (BIOS(=F1)) (DOS(=F2)) (DISK(=F3)) (CMOS(=F4))
//
// Any combination of BIOS, DOS, DISK, and CMOS can be specified.
// Filenames can be specified for each of the sections, but are 
// optional. If a section is specified without a filename, it defaults
// to 
//	<4 digit model number><section name>.BIN
//
// For example, default filenames for a TS-2200 are 2200BIOS.BIN, 
// 2200DOS.BIN, 2200DISK.BIN, and 2200CMOS.BIN.
//
// When creating the images, Flashup will NOT overwrite a file that 
// already exists. You must delete, rename, or move the old copy first.
//
// The procedure for creating update images is as follows:
// 1.	Configure a master EPC from which the clones will be copied. In 
//	particular, make sure the Flash disk contains all necessary files, 
//	and the CMOS boot options are correct.
// 2.	Install the TS-9100 / TS-9200 boot board and power up the EPC.
// 3.	Change to the directory where the update images will be stored. By
//	default, this is 'C:\IMAGES'.
// 4.	If you are replacing a previous set of images, rename or move the
//	old copies of the files. Flashup will not overwrite files.
// 4.	Run flashup, specifying the sections for which you are creating 
//	images. Specify filenames if you don't want to use the defaults.
// 5.	Power off, remove the boot board (if present) and you are done.
//
// Updating an EPC (BLAST)
// --------------------------
// To update an EPC from image files, the following syntax is used:
//
// FLASHUP /BLAST (BIOS(=F1)) (DOS(=F2)) (DISK(=F3)) (CMOS(=F4)) (RTC)
//
// Any combination of BIOS, DOS, DISK, CMOS, and RTC can be specified.
// Filenames can be specified for each of the sections, but are 
// optional. If a section is specified without a filename, it defaults
// to 
//	<4 digit model number><section name>.BIN
//
// For example, default filenames for a TS-2200 are 2200BIOS.BIN, 
// 2200DOS.BIN, 2200DISK.BIN, and 2200CMOS.BIN.
//
// The RTC option copies the time and date from an attached TS-9200 boot
// board, which contains its own RTC. Without a TS-9200, you cannot use
// this feature.
//
// The procedure for updating an EPC is as follows:
// 1.	If this is a virgin board or one on which the BIOS has been 
//	corrupted (the red LED never even blinks at poweron) then you 
//	must use a TS-9100 or TS-9200 to bootstrap the EPC to a running
//	state. Live boards can be updated directly by downloading images
//	and Flashup to the EPC and running them. In either case, Flashup
//	and the update images must be accessable from the EPC.
// 3.	Change to the directory where the update images are stored. By
//	default, this is 'C:\IMAGES'.
// 4.	Run flashup, specifying the sections you are updating. Specify
//	filenames if you are not using the defaults.
// 5.	Power off, remove the boot board (if present) and you are done.

// Revision History
// --------------------------
// 19990705 1.0	cjd	Started
// 19990901 1.1	cjd	Modified and updated to read Flash image from a file.
//			Renamed to FLASHUP
// 19990905 1.2	cjd	Added ability to write image of Flash disk to a file.
// 19990914 1.3	cjd	Added ability to selectively create or restore Flash 
//			image files
// 19990915 1.3.1	Handles 'PS' returned by Prosoft PS-4100 BIOS call
// 20000501 1.4	cjd	Cmdline improved. Filenames can now be specified 
//			on the cmdline, as well as a prefix
// 20000616 1.5	cjd	Simplified cmdline - no prefix crap.
// 20000713 1.6	cjd	Added TS-2250
// 20000718 1.7 cjd	Added CMOS and RTC options, updated docs
// 20000727 1.7.1 cjd	- Dont quit if RTC set is attempted w/o a TS9100 - so
//			we can use the same batch files on a TS9100
//			- Give better feedback (what's going to be replaced)
//			for the confirm mesg for /BLAST and /EPROM
// 20001011 1.7.2 cjd	Added /BLASTBOOT flag to write to 91/9200 flash. This
//			is VERYVERY dangerous, so it is not in the help messages..
// 20010215 1.7.3 mtj Added RTC blast for the 9300
// 20010223 1.7.4 mtj changed RTC on 9200/9300 to 0x9E 0x9F so address is 1 byte
// 20010621 1.8.0 mtj Added Omega support and format changes
// 20010709 1.8.0 mtj flash size is now determine from BIOS call
// 20010709 1.8.0 mtj removed eeprom transfer options for obsolete TS-9100
// 20010920 1.8.0 mtj read model to find flash size, add 2M flash for 5300
// 20011129 1.8.1 mtj added SetFlashSize function that uses model number
// 20011129 1.8.1 mtj added /direct option that gets addresses from command line
// 20011220 1.8.1 mtj added support for second disk on 5300
// 20020405 1.8.2 mtj added support for 3400
// 20021119 1.8.3 mtj added support for TS-5400 and TS-5500
// 20021219 1.8.4 mtj added support for TS-9400
// 20030717 1.8.5 mtj add command line option /Y to bypass prompt
// 20030825 1.9.0 mtj added support for 5600, cmos saves up to x7F on 5000 series
// 20030825 1.9.0 mtj added support for SPI eeprom to hold CMOS on 5600
// 20040126	1.9.1 mtj added support for TS-5700, same flash layout as 5600
// 20040416 1.9.2 mtj TS-9400 installed test was looking a wrong I/O address

// nuke eeprom options now that 9300 is obsolete

#define OMEGA 1

/****************************************************************************
   INCLUDE FILES
****************************************************************************/

#include <stdio.h>	// for printf
#include <stdlib.h>
#include <conio.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>
#include <iostream.h>
#include <iomanip.h>
#include <string.h>
#include <sys\stat.h>

	
#define SPI_CTRL_REG       0xCD0
#define SPI_TX_DATA        0xCD1
#define SPI_CMD_REG        0xCD2
#define SPI_STATUS_REG     0xCD3
#define SPI_RX_DATA        0xCD4

	
/****************************************************************************
	CONSTANTS
****************************************************************************/

const char VER[] = "EPC System Update Utility v1.9.2, (c) 2002 Technologic Systems, Inc.";

// CMOS defines
const int CMOS_ADDR         = 0x70;	   // address of target board RTC's latch
const int CMOS_DATA         = 0x71;	   // address of target board RTC's data
const int CMOS_ADDR_9200    = 0x9E;	   // address of boot board RTC's latch
const int CMOS_DATA_9200    = 0x9F;	   // address of boot board RTC's data

const int CMOS_YEAR     = 0x09;		   // offset of year within RTC
const int CMOS_REGA     = 0x0A;		   // offset of Register A within RTC
const int CMOS_REGB     = 0x0B;		   // offset of Register B within RTC 
const int CMOS_REGC     = 0x0C;		   // offset of Register C within RTC 
const int CMOS_REGD     = 0x0D;		   // offset of Register D within RTC 
const int CMOS_STAT     = 0x0E;		   // offset of Stat within RTC
const int CMOS_RESTART  = 0x0F;		   // restart status byte within RTC
const int CMOS_BEGIN    = 0x10;		   // start of cmos area within RTC

const int CMOS_ARRAY_SIZE      = 0x80; // array size for storing all CMOS mem.

const int LOCK  = 00;	   // commands for special flash BIOS calls
const int ERASE = 01;
const int READ  = 02;
const int WRITE = 03;

const unsigned long BOOTBASE       = 0x30000000;

const unsigned long BOOTBIOSOFFSET = 0x00070000;   // settings for 512kb flash
const unsigned long BOOTDOSOFFSET  = 0x00060000;   // on boot board (upper)
const unsigned long BOOTDISKOFFSET = 0x00000000;

const unsigned long BUFFER_SIZE  = 0x00004000;	// 16k buffer for data

const char *ErrorMesg[] = {
	"No Errors.",
	"Usage error (bad command line, or answered 'n' to erasing Flash).",
	"File read/write error.",
	"Flash read/write/erase error.",
	"Attempting to copy RTC. You must have a TS-9200 or TS-9300 or TS-9400 for this option."};


/****************************************************************************
	TYPEDEFS AND DEFINES
****************************************************************************/

#define CF 1

enum ErrorCodes {
	ERR_NONE = 0, ERR_USAGE, ERR_FILE, ERR_FLASH, ERR_NO9200
};

/****************************************************************************
	GLOBAL VARIABLES
****************************************************************************/

// These are defaulted to 2 Meg for the TS-5X00, sectors=64kb

unsigned long ulFLASHBASE   = 0x10000000;	// Flash start address, virtual
unsigned long ulDISKOFFSET  = 0x00000000;	// A: starts at beginning of flash
unsigned long ulDISKSIZE    = 0x000E0000;	// A: is E0000 bytes
unsigned long ulDOSOFFSET   = 0x00000000;	// DOS starts at E0000, 386 only
unsigned long ulBIOSOFFSET  = 0x000E0000;	// BIOS starts at E0000
unsigned long ulBIOSSIZE    = 0x00020000;	// BIOS is 20000 bytes (128kb)
unsigned long ulDISK2OFFSET = 0x10000000;	// second disk on 5300
unsigned long ulDISK2SIZE   = 0x00100000;	// second disk on 5300

int   ModelNum = 0;

/****************************************************************************
	FUNCTION PROTOTYPES
****************************************************************************/

void ErrorHandler (const int ErrNum);
void Usage ();
int CheckTSBios ();
int CheckModel();
void SetFlashSize( int Model );
unsigned char ReadCmosByte(int Base, unsigned char Address);
void ReadCmosData(int Base, unsigned char CmosData[], unsigned char Address, unsigned char Length);
void WriteCmosByte(int Base, int Address, unsigned char Data);
void WriteCmosData(int Base, unsigned char CmosData[], unsigned char Address, unsigned char Length);
int WriteCmosToFile(const char *Filename, unsigned char CmosData[]);
int WriteFileToCmos(const char *Filename, unsigned char CmosData[]);
int UpdateRtc();
int FlashCmd (int Cmd, unsigned long FlashAddr, unsigned int BufSeg, unsigned int BufOff, unsigned int Count);
int ReadFlash (unsigned long FlashAddr, unsigned int BufSeg, unsigned int BufOff, unsigned int Count);
int WriteFlash (unsigned long FlashAddr, unsigned int BufSeg, unsigned int BufOff, unsigned int Count);
int EraseFlash (unsigned long FlashAddr);
int WriteFileToFlash(const char *Filename, const unsigned long FlashAddress, const unsigned long FlashLength);
int WriteFlashToFile(const char *Filename, const unsigned long FlashAddress, const unsigned long FlashLength);
int WriteFlashToFlash(const unsigned long FlashInAddress, const unsigned long FlashOutAddress, const unsigned long FlashLength);
int main(int argc, char *argv[]);
int WriteCmosDataEeprom( unsigned char CmosData[] );
void SPI_eeprom_write_enable( void );
unsigned char SPI_transfer( unsigned char ucSendData );


/****************************************************************************
	  Function: ErrorHandler
   Description: Prints error message to stdout then exits
		Passed: index into ErrorCodes array containing messages
	   Returns: none - exits
*****************************************************************************/
void ErrorHandler (const int ErrNum) 
{
	cout << "Houston, we have a problem: " << ErrorMesg[ErrNum] << endl;
	exit (ErrNum);
}

/****************************************************************************
	  Function: Usage
   Description: Shows command line parameters and exits
		Passed: none
	   Returns: none - exits
*****************************************************************************/
void Usage ()
{
	cout << "Usage:\nFLASHUP (/BLAST|/CREATE) (BIOS(=F1)) (DOS(=F2)) (DISK(=F3)) (CMOS(=F4)) (RTC)\n\n";
	cout << "example: flashup /create bios=bios.bin dos=dos.bin disk=dsk.bin cmos=cmos.bin\n";
	cout << "example: flashup /blast bios=bios.bin dos=dos.bin disk=dsk.bin cmos=cmos.bin rtc\n";

	cout << "Choose one of the two commands:" << "\n";
	cout << "\t/BLAST\tWrite image files to Flash memory." << "\n";
	cout << "\t/CREATE\tRead Flash memory and write to image files." << "\n\n";

	cout << "Choose any combination of the following options:" << "\n";
	cout << "\tBIOS\tCreate/restore BIOS image (64KB @ 0x100F0000)" << "\n";
	cout << "\tDOS\tCreate/restore DOS image  (64KB @ 0x100E0000)" << "\n";
	cout << "\tDISK\tCreate/restore DISK image (896KB @ 0x10000000)" << "\n";
	cout << "\tCMOS\tCreate/restore CMOS image" << "\n";
	cout << "\tRTC\tCopy time and date from a TS-9200 or TS-9300 to the EPC RTC" << "\n\n";

	cout << "Optional Parameters" << "\n";
	cout << "\tF1-F4\tThe Flash image filenames can be specified directly.\n";
	cout << "\t/Y\tbypass Y/N prompt before writing to flash.\n\n";

	cout << "NOTE: When using /CREATE, Filename(s) must NOT exist - FLASHUP will NOT\n";
	cout << "      overwrite existing files." << "\n";

	cout << endl;

	exit (ERR_USAGE);
}

/****************************************************************************
	  Function: CheckTSBios
   Description: Check for TS BIOS
		Passed: none
	   Returns: 1 for TS Bios, 0 for other
*****************************************************************************/
int CheckTSBios ()
{

	union REGS regs;

	regs.x.ax = 0xB000;				// TS BIOS info function
	int86 (0x15, &regs, &regs);	// do it.

	// 'TS' should be returned in DX (the PS-4100 returns PS, so handle that too..)
	if( regs.h.ah == 0 && ((regs.x.dx == 0x5453) || (regs.x.dx == 0x5053)) )
		return(regs.h.cl); // YES - return the model number
	else
		return(0);		   // NO this is NOT TS BIOS
}

/****************************************************************************
	  Function: CheckModel
   Description: calls TS BIOS function and prints out model #
		Passed: none
	   Returns: model number, or EXITS if not a TS BIOS
*****************************************************************************/
int CheckModel()
{
	int   Model;

	cout << "Model number: ";
	Model = CheckTSBios();
	cout << "[" << Model << "] ";
	switch( Model ) {
		case 0:
			cout << "Not a TS EPC! This PC is not supported - exiting." << endl;
			exit (ERR_USAGE);
			break;
		case 21:    cout << "TS-2100 EPC" << endl; return(2100);	//break;
		case 22:    cout << "TS-2200 EPC" << endl; return(2200);	//break;
		case 23:    cout << "TS-2350 EPC" << endl; return(2350);	//break;
		case 24:    cout << "TS-2250 EPC" << endl; return(2250);	//break;
		case 28:    cout << "TS-2800 EPC" << endl; return(2800);	//break;
		case 31:    cout << "TS-3100 EPC" << endl; return(3100);	//break;
		case 32:    cout << "TS-3200 EPC" << endl; return(3200);	//break;
		case 33:    cout << "TS-3300 EPC" << endl; return(3300);	//break;
		case 34:    cout << "TS-3400 EPC" << endl; return(3400);	//break;
		case 53:    cout << "TS-5300 EPC" << endl; return(5300);	//break;
		case 54:    cout << "TS-5400 EPC" << endl; return(5400);	//break;
		case 55:    cout << "TS-5500 EPC" << endl; return(5500);	//break;
		case 56:    cout << "TS-5600 EPC" << endl; return(5600);	//break;
		case 57:    cout << "TS-5700 EPC" << endl; return(5700);	//break;
		case 41:    cout << "PS-4100 EPC" << endl; return(4100);	//break;
		default:
			cout << "Unknown TS EPC model! (" << Model << ") - continuing" << endl; 
			return(Model);
			//break;
	}
	return(Model);
}

/****************************************************************************
	  Function: SetFlashSize
   Description: Uses model number to determine flash offset and length
		Passed: int Model number detected
	   Returns: none
*****************************************************************************/
void SetFlashSize( int Model )
{

	cout << "SetSize" << Model << "\n";

	switch( Model ) {
		// smart card
		case 3100:
			ulFLASHBASE     = 0x10000000;
			ulDISKOFFSET    = 0x00000000;
			ulDISKSIZE      = 0x00060000;
			ulDOSOFFSET     = 0x00060000;
			ulBIOSOFFSET    = 0x00070000;
			ulBIOSSIZE      = 0x00010000;
			break;
		case 2100:
		case 2200:
		case 2300:
		case 2400:
		case 2800:
		case 3200:
		case 3300:
		case 3400:
		case 4100:
			ulFLASHBASE     = 0x10000000;
			ulDISKOFFSET    = 0x00000000;
			ulDISKSIZE      = 0x000E0000;
			ulDOSOFFSET     = 0x000E0000;
			ulBIOSOFFSET    = 0x000F0000;
			ulBIOSSIZE      = 0x00010000;
			break;
		case 5300:
		case 5400:
		case 5500:
			ulFLASHBASE     = 0x10000000;
			ulDISKOFFSET    = 0x00000000;
			ulDISKSIZE      = 0x000E0000;
			ulDOSOFFSET     = 0x00000000;
			ulBIOSOFFSET    = 0x000E0000;
			ulBIOSSIZE      = 0x00020000;
			ulDISK2OFFSET   = 0x10000000;	// second disk on 5300
			ulDISK2SIZE     = 0x00100000;	// second disk on 5300
			break;
		case 5600:
		case 5700:
			ulFLASHBASE     = 0x10000000;
			ulDISKOFFSET    = 0x00000000;
			ulDISKSIZE      = 0x001E0000;
			ulDOSOFFSET     = 0x00000000;
			ulBIOSOFFSET    = 0x001E0000;
			ulBIOSSIZE      = 0x00020000;
			break;
		default:
			cout << "Model number" << Model << " not found\n";
			exit (ERR_FLASH);

	}
}

/****************************************************************************
	  Function: ReadCmosByte
   Description: Read a byte from CMOS memory in RTC
		Passed: Base address of RTC, offset within RTC
	   Returns: Byte read from CMOS
*****************************************************************************/
unsigned char ReadCmosByte(int Base, unsigned char Address)
{
	unsigned char  tmp = 0;
	int   i = 0;

	asm   cli;			// turn off interrupts

	if( Address <= CMOS_YEAR ) {
		// if the RTC is updating, wait until it is finished
		for( outportb(Base, CMOS_REGA); (inportb(Base + 1) & 0x80); outportb(Base, CMOS_REGA) ) {
			i++;
		}
	}

	outportb(Base, Address);
	tmp = inportb(Base + 1);

	if( Address <= CMOS_YEAR ) {
		// if the RTC started updating while we were reading, wait until it is finished and try again
		for( outportb(Base, CMOS_REGA); (inportb(Base + 1) & 0x80); outportb(Base, CMOS_REGA) ) {
			i++;
		}

		outportb(Base, Address);
		tmp = inportb(Base + 1);
	}

	asm   sti;			// turn on interrupts
	return(tmp);
}

/****************************************************************************
	  Function: ReadCmosData
   Description: Reads contiguous bytes from CMOS into an array
		Passed: Base addr of RTC, array ptr, start offset within RTC, length
	   Returns: none
*****************************************************************************/
void ReadCmosData(int Base, unsigned char CmosData[], unsigned char Address, unsigned char Length)
{
	unsigned char  i;

	cout << "Reading " << hex << (int)Length << "h bytes at cmos address " << (int)Address << "h\n";

	for( i = Address; i < (Address + Length); i++ ) {
		CmosData[i] = ReadCmosByte(Base, i);
	}
}

/****************************************************************************
	  Function: WriteCmosByte
   Description: writes byte into CMOS memory in RTC
		Passed: Base addr of RTC, offset within RTC, byte to write
	   Returns: none
*****************************************************************************/
void WriteCmosByte(int Base, int Address, unsigned char Data)
{
	asm   cli;

	// if this is a RTC register, turn off updates while we are in here
	if( Address <= CMOS_YEAR ) {
		outportb(Base, CMOS_REGB);
		outportb(Base + 1, (inportb(Base + 1) | 0x80));
	}

	outportb(Base, Address);
	outportb(Base + 1, Data);

	// if this is a RTC register, turn updates back on 
	if( Address <= CMOS_YEAR ) {
		outportb(Base, CMOS_REGB);
		outportb(Base + 1, (inportb(Base + 1) & 0x7F));
	}

	asm   sti;
}

/****************************************************************************
	  Function: WriteCmosData
   Description: Write bytes from an array into CMOS memory in RTC
		Passed: Base addr of RTC, array of data to write, offset within RTC, len
	   Returns: none
*****************************************************************************/
// unsigned char WriteCmosData()
// 
//	Reads bytes from an array into Cmos
//
//	Base		base i/o address of RTC
// 	CmosData	array to store data in
//	Address		RTC address at which to start 
//	Length		number of bytes to read
//
//
void WriteCmosData(int Base, unsigned char CmosData[], unsigned char Address, unsigned char Length)
{
	unsigned char  i;

	cout << "Writing " << hex << (int)Length << "h bytes at cmos address " << (int)Address << "h\n";
	for( i = Address; i < (Address + Length); i++ ) {
		WriteCmosByte(Base, i, CmosData[i]);
	}
}

// dumps the contents of the CMOS and RTC to the console
int DumpCmos(unsigned char CmosData[])
{
	int   i;		// loop control

	cout << "Dumping CMOS Array:\n" << hex;
	for( i = 0; i < CMOS_ARRAY_SIZE; i++ ) {
		cout << setw(4) << (int)CmosData[i];
		if( (i + 1) % 16 == 0 )
			cout << "\n";

	}
	cout << endl;
	return(0);
}

/****************************************************************************
	  Function: WriteCmosToFile
   Description: Read CMOS data then write array to binary file
		Passed: Filename to save as, array for CMOS data
	   Returns: 0 if no error, error code if error
*****************************************************************************/
int WriteCmosToFile(const char *Filename, unsigned char CmosData[])
{

	int   out_file;		// output file descriptor
	unsigned long  BytesWritten;	 // number of bytes written to the file

	out_file = open(Filename, O_WRONLY | O_CREAT | O_EXCL | O_BINARY, S_IWRITE);	// attempt to open new Filename - error if it already exists 
	if( out_file < 0 ) {
		cerr << "Error: Cannot open " << Filename << ". Does it already exist?" << endl;
		return(ERR_FILE);
	}

	ReadCmosData(CMOS_ADDR, CmosData, CMOS_BEGIN, (CMOS_ARRAY_SIZE - CMOS_BEGIN));

	if( (BytesWritten = write(out_file, CmosData, CMOS_ARRAY_SIZE)) != CMOS_ARRAY_SIZE ) {
		cerr << "Error: cannot write to file (disk full?)" << endl;
		return(ERR_FILE);
	}

	// All finished. Close things down and report success
	cout << "\rReading Cmos memory: complete                              " << endl; 
	close(out_file);
	cout << "Flash image file successfully updated. Total bytes = 0x" << hex << BytesWritten << endl;

	return(ERR_NONE);						 // successful exit
}     

/****************************************************************************
	  Function: WriteFileToCmos
   Description: Read CMOS binary file and write dato to CMOS memory
		Passed: Filename to read, array for CMOS data
	   Returns: 0 if no error, error code if error
*****************************************************************************/
int WriteFileToCmos(const char *Filename, unsigned char CmosData[])
{

	int   in_file;		// input file descriptor
	unsigned long  read_size = 0;	// bytes read in last read

	in_file = open(Filename, O_RDONLY | O_BINARY);	// open the input file, binary read only
	if( in_file < 0 ) {
		cerr << "Error: cannot open " << Filename << endl;
		return(ERR_FILE);
	}

	if( (read_size = read(in_file, CmosData, CMOS_ARRAY_SIZE)) != CMOS_ARRAY_SIZE ) {
		cerr << "Error: cannot read from file (is it empty?)" << endl;
		return(ERR_FILE);
	}

	if( ModelNum < 5000 ) {		// don't write over MAC address in CMOS
		WriteCmosData(CMOS_ADDR, CmosData, CMOS_BEGIN, 0x40-CMOS_BEGIN);
	} else if( ModelNum < 5600 ) {	// custom CMOS goes to 0x46, save up to 0x7F
		WriteCmosData(CMOS_ADDR, CmosData, CMOS_BEGIN, 0x7F-CMOS_BEGIN);
	} else {					// must be a 5600, save to SPI eeprom
		WriteCmosDataEeprom( CmosData );
		WriteCmosData(CMOS_ADDR, CmosData, CMOS_BEGIN, 0x7F-CMOS_BEGIN);
	}

	// All finished. Close things down and report success
	cout << "\rWriting Cmos memory: complete                              " << endl; 
	close(in_file);
	cout << "Cmos memory successfully updated. Total bytes = 0x" << hex << read_size << endl;

	return(ERR_NONE);						 // successful exit
}     

/****************************************************************************
	  Function: UpdateRtc
   Description: Read RTC from blast board, write time and date to target board
		Passed: none
	   Returns: 0 if no error, error code if error
*****************************************************************************/
int UpdateRtc()
{

	unsigned char  RtcData[CMOS_ARRAY_SIZE] = {0xFF};

	cout << "Checking for presence of TS-9200 or TS-9300 or TS-9400... ";
	if( (((inportb(0x475) & 0x07) != 0x04) && 
		 (inportb(0x475) != 0x32)) && 
		(inportb(0x19D) != 0x3A) &&
		(inportb(0x18D) != 94) ) {
		cout << "not found. Skipping RTC set.\n";
		return(ERR_NO9200);
	} else {
		cout << "found aok.\n";
	}

	// CMOS century value is at address 32h in the RTC, this doesn't get copied 
	// in the following command, and will cause a y2k bug that shows the date as
	// 1901 instead of 2001 when viewed from CMOS setup. when you blast the CMOS
	// this value gets copied, then the y2k bug will be fixed.

	// get RTC data from TS-9200
	ReadCmosData(CMOS_ADDR_9200, RtcData, 0, 0x0B);

	// copy to base board
	WriteCmosData(CMOS_ADDR, RtcData, 0, 0x0B);

	// All finished. Close things down and report success
	cout << "\rWriting RTC: complete                              " << endl; 
	cout << "RTC successfully updated." << endl;

	return(ERR_NONE);						 // successful exit
}     

/****************************************************************************
	  Function: FlashCmd
   Description: calls special BIOS function for flash: erase / read / write
		Passed: flash command, flash address, buffer segment, offset, count
	   Returns: 0 if no error, 1 if error
*****************************************************************************/
int FlashCmd (int Cmd, unsigned long FlashAddr, unsigned int BufSeg, unsigned int BufOff, unsigned int Count)
{
	struct REGPACK reg;

	reg.r_ax = ((0xE0 << 8) | (Cmd & 0x00FF));	// shift E0h into  AH, put Cmd in AL

	reg.r_di = FlashAddr >> 16;		// shift FlashAddr MSW into DI
	reg.r_si = FlashAddr & 0x0000FFFF;	// Mask FlashAddr LSW, place in si
	reg.r_cx = Count;			// Load number of bytes to work on
	reg.r_es = BufSeg;
	reg.r_bx = BufOff;


//	cout << "\nFlashCmd: cmd=" << reg.r_ax << " Flashaddr=" << reg.r_di << ":" << reg.r_si 
//			<< " cnt=" << reg.r_cx << " Bufaddr=" << reg.r_es << ":" << reg.r_bx 
//			<< " flags=" << reg.r_flags << endl;

	intr(0x15, &reg);			// do it	

//	cout << "\nFlashCmd: cmd=" << reg.r_ax << " Flashaddr=" << reg.r_di << ":" << reg.r_si 
//			<< " cnt=" << reg.r_cx << " Bufaddr=" << reg.r_es << ":" << reg.r_bx 
//			<< " flags=" << reg.r_flags << endl;

	if( reg.r_flags & CF ) {
//		cout << "Flashcmd: Returning ERROR!\n";
		return(1);	  // Flash command failure	
	} else {
//		cout << "Flashcmd: Returning AOK!\n";		
		return(0);	  // Flash command success
	}

} 


/****************************************************************************
	  Function: Readflash
   Description: calls FlashCmd
		Passed: flash address, buffer segment, buffer offset, count
	   Returns: always 0
*****************************************************************************/
int ReadFlash (unsigned long FlashAddr, unsigned int BufSeg, unsigned int BufOff, unsigned int Count)
{
//	return (FlashCmd (READ, FlashAddr, BufSeg, BufOff, Count));
	FlashCmd (READ, FlashAddr, BufSeg, BufOff, Count);
	return(0);					  // Always get errors with BIOS call - WHY?

}

/****************************************************************************
	  Function: WriteFlash
   Description: calls FlashCmd
		Passed: flash address, buffer segment, buffer offset, count
	   Returns: 0 if no error, 1 if error
*****************************************************************************/
int WriteFlash (unsigned long FlashAddr, unsigned int BufSeg, unsigned int BufOff, unsigned int Count)
{
//	// saftey check - prevents us from accidentally erasing the DOS or BIOS images.
//	if ((FlashAddr + Count - 1) >= 0x100E0000)
//		return(2);
//	else
	return(FlashCmd (WRITE, FlashAddr, BufSeg, BufOff, Count));
}

/****************************************************************************
	  Function: EraseFlash
   Description: calls FlashCmd
		Passed: address to erase
	   Returns: 0 if no error, 1 if error
*****************************************************************************/
int EraseFlash (unsigned long FlashAddr)
{
//	// saftey check - prevents us from accidentally erasing the DOS or BIOS images.
//	if (FlashAddr >= 0x100E0000)
//		return(2);
//	else
	return(FlashCmd (ERASE, FlashAddr, 0, 0, 0));
}

/****************************************************************************
	  Function: WriteFileToFlash
   Description: Read file, erase flash, write to flash
		Passed: file name, flash address, length of flash to write
	   Returns: 0 if no error, error code if error
*****************************************************************************/
int WriteFileToFlash(const char *Filename, const unsigned long FlashAddress, const unsigned long FlashLength)
{
	char  Buffer[BUFFER_SIZE];	// data buffer
//	int 	i = 0;			// loop count
	int   in_file;		// input file descriptor
	unsigned long  read_size = 0;	// bytes read in last read
	unsigned long  read_total = 0;	// running total of bytes
	unsigned long  FlashTemp = 0;	// various temporary large numbers


	in_file = open(Filename, O_RDONLY | O_BINARY);	// open the input file, binary read only
	if( in_file < 0 ) {
		cerr << "Error: cannot open " << Filename << endl;
		return(ERR_FILE);
	}
	FlashTemp = filelength(in_file);	 // small sanity check here 
//	cout << "Image file size: 0x" << hex << FlashTemp << endl;
	if( FlashTemp < FlashLength ) {
		cout << "Warning: Image file is smaller than intended Flash memory section. Continuing.\n";
	}

	// erase the Flash memory
	for( FlashTemp = FlashAddress; FlashTemp < (FlashAddress + FlashLength); FlashTemp += 0x10000 ) {

		cout << "\rErasing Flash memory: Address = 0x" << hex << FlashTemp;

		if( (EraseFlash (FlashTemp)) != 0 ) {
			cout << endl << endl << "Flash memory erase command failure, exiting." << endl;
			return(ERR_FLASH);
		}
	}
	cout << "\rErasing Flash memory: complete                           " << endl;

	// write the new Flash contents
	read_total = 0;

	for( FlashTemp = FlashAddress; 
	   (read_size = read(in_file, Buffer, sizeof(Buffer))) != 0; 
	   FlashTemp += read_size ) {

		cout << "\rWriting Flash memory: address = 0x" << hex << FlashTemp
		<< ", Bytes = 0x" << hex << read_size;

		if( (WriteFlash(FlashTemp, FP_SEG(Buffer), FP_OFF(Buffer), read_size)) != 0 ) {
			cout << endl << endl << "Flash memory write command failure, exiting." << endl;
			return(ERR_FLASH);
		}
		read_total += read_size;  // increment bytes read count
	}

	// All finished. Close things down and report success
	close(in_file);
	cout << "\rWriting Flash memory: Done. Total bytes updated = 0x" << hex << read_total << endl;

	return(ERR_NONE);						 // successful exit
}

/****************************************************************************
	  Function: WriteFlashToFile
   Description: Open file, read flash, save to file
		Passed: file name, flash address, length to read
	   Returns: 0 if no error, error code if error
*****************************************************************************/
int WriteFlashToFile(const char *Filename, const unsigned long FlashAddress, const unsigned long FlashLength)
{
	char  Buffer[BUFFER_SIZE];	// data buffer
//	int 	i = 0;			// loop count
	int   out_file;		// output file descriptor
	unsigned long  BytesWritten;	 // number of bytes written to the file
	unsigned long  BytesTotal = 0;	// total bytes written to the file
	unsigned long  FlashTemp = 0;	// various temp. large numbers

	out_file = open(Filename, O_WRONLY | O_CREAT | O_EXCL | O_BINARY, S_IWRITE);	// attempt to open new Filename - error if it already exists 
	if( out_file < 0 ) {
		cerr << "Error: Cannot open " << Filename << ". Does it already exist?" << endl;
		return(ERR_FILE);
	}

	for( FlashTemp = FlashAddress; FlashTemp < (FlashAddress + FlashLength); FlashTemp += BUFFER_SIZE ) {
		cout << "\rReading Flash memory: address = 0x" << hex << FlashTemp
		<< ", Bytes = 0x" << hex << BUFFER_SIZE;

		if( ReadFlash (FlashTemp, FP_SEG(Buffer), FP_OFF(Buffer), BUFFER_SIZE) ) {
			cerr << "Error: cannot read from Flash memory - Unknown Error" << endl;
			return(ERR_FLASH);
		}
		if( (BytesWritten = write(out_file, Buffer, BUFFER_SIZE)) != BUFFER_SIZE ) {
			cerr << "Error: cannot write to file (disk full?)" << endl;
			return(ERR_FILE);
		}
		BytesTotal += BytesWritten;
	}

	// All finished. Close things down and report success
	cout << "\rReading Flash memory: complete                              " << endl; 
	close(out_file);
	cout << "Flash image file successfully updated. Total bytes = 0x" << hex << BytesTotal << endl;

	return(ERR_NONE);						 // successful exit
}     

/****************************************************************************
	  Function: main
   Description: startup code and command line parsing to determine actions
		Passed: command line
	   Returns: 0 if no error, error code if error
*****************************************************************************/
int main(int argc, char *argv[])
{
	char  Flag;						// confirmation response
	int   i;						// temp loop control
	int   Error = 0;				// Error code holder
	char  StrTmp[256] = "";			// holder for cmd line arguments
	char  FileDisk[20] = "";		// filename holder
	char  FileDisk2[20] = "";		// filename holder
	char  FileDos[20] = "";			// filename holder
	char  FileBios[20] = "";		// filename holder
	char  FileCmos[20] = "";		// filename holder
	char  *Ptr;
	int   DiskFlag = 0;				// read/write disk image
	int   Disk2Flag = 0;			// read/write disk image
	int   DosFlag = 0;				// r/w dos image
	int   BiosFlag = 0;				// r/w bios image
	int   CmosFlag = 0;				// r/w bios image
	int   RtcFlag = 0;				// update RTC from TS-9200
	int   WriteBoot = 0;			// write TO Bootboard (TS-91/9200) if set
	int   WriteFlash = 0;			// write TO Flash if set
	int   WriteFile = 0;			// write TO image files if set
	char  FilePrefix[8] = "";	// filename prefix (optional)
	unsigned char CmosData[128] = {0xFF};
	int Prompt = 1;					// clear if /Y on command line

	// added 20011129 mtj for direct flash mode which uses command line 
	// parameters for flash location and size.

	unsigned long FlashStart = 0x0;
	unsigned long FlashSize  = 0x0;
	char FileName[20] = "";


	cout << VER << endl << endl;	// print program version

	if( argc < 2 )
		Usage();		// invalid command line - display usage and exit

// ****************************************************************************
// ****************************************************************************

	if( !stricmp( argv[1], "/DIRECT") || !stricmp( argv[1],"-DIRECT") ) {

		if( argc < 6 )
			Usage();		// invalid command line - display usage and exit

		// CHECK FOR TS BIOS, THIS BABY COULD BE DANGEROUS

		FlashStart = strtoul( argv[3], &Ptr, 16 );
		FlashSize  = strtoul( argv[4], &Ptr, 16 );

		strncpy( FileName, argv[5], 12);

//		cout << "Start " << FlashStart << " Size " << FlashSize << " Filename " << FileName << "\n"; 
		printf( "Start: (%s) = (%lu) \n", argv[3], FlashStart );
		printf( "Size : (%s) = (%lu) \n", argv[4], FlashSize );
		printf( "Name : (%s) = (%s)  \n", argv[5], FileName );

		if( !stricmp( argv[2], "/BLAST") || !stricmp( argv[2],"-BLAST") ) {

			cout << "Writing binary image file " << FileName << " to Flash memory:\n";
			Error = WriteFileToFlash( FileName, FlashStart, FlashSize );
			if( Error )
				ErrorHandler( Error );
		} else if( !stricmp( argv[2], "/CREATE") || !stricmp( argv[2],"-CREATE") ) {

			cout << "Writing flash image to binary file " << FileName << ":\n";
			Error = WriteFlashToFile( FileName, FlashStart, FlashSize );
			if( Error )
				ErrorHandler(Error);
		} else
			Usage();

		return(ERR_NONE);
	}

// ****************************************************************************
// ****************************************************************************

	ModelNum = CheckModel();		  // find and display EPC model number

	for( i = 1; i < argc ; ++i ) {
		strncpy(StrTmp, argv[i], 255);

		if( !stricmp( StrTmp, "/BLAST") || !stricmp( StrTmp,"-BLAST") ) {
			WriteFlash = 1;

		} else if( !stricmp( StrTmp, "/CREATE") || !stricmp( StrTmp,"-CREATE") ) {
			WriteFile = 1;

		} else if( !stricmp( StrTmp, "/BLASTBOOT") || !stricmp( StrTmp,"-BLASTBOOT") ) {
			WriteBoot = 1;

		} else if( !strncmpi(StrTmp, "DISK", 4) ) {
			DiskFlag = 1;
			if( (StrTmp[4] == '=') && (StrTmp[5] != '\0') ) {
				strncpy(FileDisk, &StrTmp[5], 12);
				FileDisk[13] = '\0';
			}

		} else if( !strncmpi(StrTmp, "2DISK", 5) ) {
			Disk2Flag = 1;
			if( (StrTmp[5] == '=') && (StrTmp[6] != '\0') ) {
				strncpy(FileDisk2, &StrTmp[6], 12);
				FileDisk2[13] = '\0';
			}

		} else if( !strncmpi(StrTmp, "DOS", 3) ) {
			DosFlag = 1;
			if( (StrTmp[3] == '=') && (StrTmp[4] != '\0') ) {
				strncpy(FileDos, &StrTmp[4], 12);
				FileDisk[13] = '\0';
			}

		} else if( !strncmpi(StrTmp, "BIOS", 4) ) {
			BiosFlag = 1;
			if( (StrTmp[4] == '=') && (StrTmp[5] != '\0') ) {
				strncpy(FileBios, &StrTmp[5], 12);
				FileDisk[13] = '\0';
			}

		} else if( !strncmpi(StrTmp, "CMOS", 4) ) {
			CmosFlag = 1;
			if( (StrTmp[4] == '=') && (StrTmp[5] != '\0') ) {
				strncpy(FileCmos, &StrTmp[5], 12);
				FileCmos[13] = '\0';
			}

		} else if( !strncmpi(StrTmp, "RTC", 3) ) {
			RtcFlag = 1;

		} else if( !strncmpi(StrTmp, "ALL", 3) ) {
			DiskFlag = 1;
			if( ModelNum == 5300 )
				Disk2Flag = 1;
			DosFlag = 1;
			BiosFlag = 1;
			CmosFlag = 1;
			RtcFlag = 1;

		} else if( !strncmpi(StrTmp, "/Y", 3) ) {
			Prompt = 0;
		} else {
			Usage();							// bad command flag / help
		}
	}

	itoa(ModelNum, FilePrefix, 10);

	if( FileDisk[0] == '\0' ) {
		strcpy(FileDisk, FilePrefix);
		strncat(FileDisk, "DISK.BIN", 11);
	}
	if( FileDisk2[0] == '\0' ) {
		strcpy(FileDisk2, FilePrefix);
		strncat(FileDisk2, "2DSK.BIN", 11);
	}
	if( FileDos[0] == '\0' ) {
		strcpy(FileDos, FilePrefix);
		strncat(FileDos, "DOS.BIN", 11);
	}
	if( FileBios[0] == '\0' ) {
		strcpy(FileBios, FilePrefix);
		strncat(FileBios, "BIOS.BIN", 11);

	}
	if( FileCmos[0] == '\0' ) {
		strcpy(FileCmos, FilePrefix);
		strncat(FileCmos, "CMOS.BIN", 11);

	}


	outp (0x472, 0x01);				  // RED LED ON
	outp (0x473, 0x00);				  // GREEN LED OFF


	SetFlashSize( ModelNum );


	if( WriteFlash ) {
		if( Prompt ) {
			cout << "Preparing to update the EPC." << endl;
			cout << "WARNING: ALL CURRENT CONTENTS OF THE EPC's: ";
			if( BiosFlag ) cout << "BIOS ";
			if( DosFlag ) cout << "DOS ";
			if( DiskFlag ) cout << "FLASHDISK ";
			if( CmosFlag ) cout << "CMOS ";
			if( RtcFlag ) cout << "RTC ";
			cout << " WILL BE LOST!" << endl << endl;
			cout << "Continue? (Y/N) ";
			Flag = getch();				// get response. getche() doesn't appear to work..
			if( Flag != 'Y' && Flag != 'y' ) {
				exit(ERR_USAGE);
			}
		}

		// CMOS stuff
		if( CmosFlag ) {
			cout << "Writing file to EPC CMOS: " << FileCmos << endl;
			Error = WriteFileToCmos(FileCmos, CmosData);
			if( Error )	ErrorHandler(Error);
		}

		// RTC stuff
		//
		// Make sure we are running on a TS-9200 or TS-9300 for RTC update to work
		if( RtcFlag ) {
			cout << "Updating RTC contents: " << endl;
			Error = UpdateRtc();
			if( Error && (Error != ERR_NO9200) ) ErrorHandler(Error);
		}

		if( BiosFlag ) {
			cout << "Writing BIOS image file " << FileBios << " to Flash memory:\n";
			Error = WriteFileToFlash(FileBios, ulFLASHBASE + ulBIOSOFFSET, ulBIOSSIZE);
			if( Error )	ErrorHandler(Error);
		}
		if( DosFlag ) {
			cout << "Writing DOS image file " << FileDos << " to Flash memory:\n";
			Error = WriteFileToFlash(FileDos,  ulFLASHBASE + ulDOSOFFSET,  0x10000);
			if( Error )	ErrorHandler(Error);
		}
		if( DiskFlag ) {
			cout << "Writing disk image file " << FileDisk << " to Flash memory:\n";
			Error = WriteFileToFlash(FileDisk, ulFLASHBASE + ulDISKOFFSET, ulDISKSIZE);	// mtj 20010621
			if( Error )	ErrorHandler(Error);
		}
		if( Disk2Flag ) {
			cout << "Writing disk 2 image file " << FileDisk2 << " to Flash memory:\n";
			Error = WriteFileToFlash(FileDisk2, ulFLASHBASE + ulDISK2OFFSET, ulDISK2SIZE); // mtj 20011220
			if( Error )	ErrorHandler(Error);
		}

	}											  // if (WriteFlash)
	else if( WriteFile ) {
		// CMOS stuff
		if( CmosFlag ) {
			cout << "Writing CMOS to file " << FileCmos << endl;
			Error = WriteCmosToFile(FileCmos, CmosData);
			if( Error )	ErrorHandler(Error);
		}

		if( BiosFlag ) {
			cout << "Writing BIOS to image file " << FileBios << ":\n";
			Error = WriteFlashToFile(FileBios, ulFLASHBASE + ulBIOSOFFSET, ulBIOSSIZE);
			if( Error )	ErrorHandler(Error);
		}
		if( DosFlag ) {
			cout << "Writing DOS to image file " << FileDos << ":\n";
			Error = WriteFlashToFile(FileDos,  ulFLASHBASE + ulDOSOFFSET,  0x10000);
			if( Error )	ErrorHandler(Error);
		}
		if( DiskFlag ) {
			cout << "Writing disk image to file " << FileDisk << ":\n";
			Error = WriteFlashToFile(FileDisk, ulFLASHBASE + ulDISKOFFSET, ulDISKSIZE);
			if( Error )	ErrorHandler(Error);
		}
		if( Disk2Flag ) {
			cout << "Writing disk 2 image to file " << FileDisk2 << ":\n";
			Error = WriteFlashToFile(FileDisk2, ulFLASHBASE + ulDISK2OFFSET, ulDISK2SIZE);
			if( Error )	ErrorHandler(Error);
		}

	}								 // if (WriteFile)

	else if( WriteBoot ) {
		cout << "Preparing to update the BOOT BOARD FLASH" << endl;
		cout << "WARNING: CURRENT CONTENTS OF THE BOOT BOARD'S: ";
		if( BiosFlag ) cout << "BIOS ";
		if( DosFlag ) cout << "DOS ";
		if( DiskFlag ) cout << "FLASHDISK ";
		cout << " WILL BE LOST!" << endl << endl;
		cout << "Continue? (Y/N) ";
		Flag = getch();				// get response. getche() doesn't appear to work..

		if( Flag != 'Y' && Flag != 'y' ) {
			exit(ERR_USAGE);
		}

		if( BiosFlag ) {
			cout << "Writing BIOS image file " << FileBios << " to Flash memory:\n";
			Error = WriteFileToFlash(FileBios, BOOTBASE + BOOTBIOSOFFSET, ulBIOSSIZE);
			if( Error )	ErrorHandler(Error);
		}
		if( DosFlag ) {
			cout << "Writing DOS image file " << FileDos << " to Flash memory:\n";
			Error = WriteFileToFlash(FileDos,  BOOTBASE + BOOTDOSOFFSET,  0x10000);
			if( Error )	ErrorHandler(Error);
		}
		if( DiskFlag ) {
			cout << "Writing disk image file " << FileDisk << " to Flash memory:\n";
			Error = WriteFileToFlash(FileDisk, BOOTBASE + BOOTDISKOFFSET, 0x60000);
			if( Error )	ErrorHandler(Error);
		}

	}	 // if (WriteFlash)

	outp (0x472, 0x00);				  // RED LED OFF
	outp (0x473, 0x01);				  // GREEN LED ON

	if( WriteFlash ) {
		cout << "\nDone. Turn off power, remove the boot board, and turn the power back on." << endl;
	}

	return(ERR_NONE);
}
/****************************************************************************
*****************************************************************************/

/****************************************************************************
	  Function:
   Description:
		Passed:
	   Returns:
*****************************************************************************/

int WriteCmosDataEeprom( unsigned char CmosData[] )
{
	unsigned char ucStatus;
	int MMCR_segment = 0xDF00;
	unsigned char far *myloc;
	unsigned char ucTemp1;
	int i,j;
	unsigned char error = 0;


	// setup control register
	myloc = (unsigned char far *) MK_FP ( MMCR_segment, SPI_CTRL_REG );
	*myloc = 0x37;


	// write enable entire eeprom ============================================


	// eeprom CS on
	ucTemp1 = inportb( 0x76 );
	outportb( 0x76, ucTemp1 | 0x40 );

	// send write status register command
	SPI_transfer( 0x01 );

	// send write enable for entire eeprom
	SPI_transfer( 0x00 );

	// eeprom CS off
	ucTemp1 = inportb( 0x76 );
	outportb( 0x76, ucTemp1 & 0xBF );

	delay( 1 );


	// write data to eeprom, one sixteen byte page at a time.=================

	for( i=1; i<8; i++ ) {

		// must write enable each page
		SPI_eeprom_write_enable();

		// eeprom CS on
		ucTemp1 = inportb( 0x76 );
		outportb( 0x76, ucTemp1 | 0x40 );

		// send write data command
		SPI_transfer( 0x02 );

		// send address
		SPI_transfer( (unsigned char) (i*16) );

		for( j=0; j<16; j++ ) {

			// send data
			SPI_transfer( (unsigned char) *(CmosData+(i*16+j)) );
		}

		// eeprom CS off
		ucTemp1 = inportb( 0x76 );
		outportb( 0x76, ucTemp1 & 0xBF );

		delay(1);

		// eeprom CS on
		ucTemp1 = inportb( 0x76 );
		outportb( 0x76, ucTemp1 | 0x40 );

		// read status reg to check busy bit
		SPI_transfer( 0x05 );

		for( j=0; j<100; j++ ) {

			ucTemp1 = SPI_transfer( 0xFF );

			// done if clear
			if( (ucTemp1 & 0x01) == 0x00 )
				break;

			delay( 1 );
		}

		// eeprom CS off
		ucTemp1 = inportb( 0x76 );
		outportb( 0x76, ucTemp1 & 0xBF );

		if( j>99 ) {
			cout << "Time out waiting for busy bit to clear\n" ;
			error = 1;
		}
	}
return error;

}

/****************************************************************************
	  Function:
   Description:
		Passed:
	   Returns:
*****************************************************************************/
void SPI_eeprom_write_enable( void )
{
unsigned char ucTemp1;
	
	// eeprom CS on
	ucTemp1 = inportb( 0x76 );
	outportb( 0x76, ucTemp1 | 0x40 );

	// send write enable command
	SPI_transfer( 0x06 );

	// eeprom CS off
	ucTemp1 = inportb( 0x76 );
	outportb( 0x76, ucTemp1 & 0xBF );

	delay( 1 );
}

/****************************************************************************
	  Function:
   Description:
		Passed:
	   Returns:
*****************************************************************************/
unsigned char SPI_transfer( unsigned char ucSendData )
{
	
	unsigned char ucStatus;
	int MMCR_segment = 0xDF00;
	unsigned char far *loc;


	// setup transmit data
	loc = (unsigned char far *) MK_FP ( MMCR_segment, SPI_TX_DATA );
	*loc = ucSendData;

	// start transaction
	loc = (unsigned char far *) MK_FP ( MMCR_segment, SPI_CMD_REG );
	*loc = 0x03;

	// wait for busy bit to clear
	do {
		loc = (unsigned char far *) MK_FP ( MMCR_segment, SPI_STATUS_REG );
		ucStatus = *loc;
	} while( (ucStatus & 0x02) != 0x00 );

	// read data
	loc = (unsigned char far *) MK_FP ( MMCR_segment, SPI_RX_DATA );

	return *loc;
}


